{-# LANGUAGE InstanceSigs #-}
{-# LANGUAGE TypeOperators #-}
module Language.Java.Classfile.Extractable () where
import Language.Java.Classfile.Extract (Extract, bytes)
import Data.Word (Word8, Word16, Word32)
import qualified Data.ByteString.Lazy as ByteString
import Data.ByteString.Lazy (ByteString)
import Data.Bits (Bits(shiftL, (.|.)))
import GHC.Generics (U1 (U1), (:*:))
import Control.Applicative ((<|>))


class Extractable a where
  extract :: Extract a

instance Extractable Word8 where
  extract :: Extract Word8
  extract = ByteString.head <$> bytes 1

instance Extractable Word16 where
  extract :: Extract Word16
  extract = build <$> bytes 2

instance Extractable Word32 where
  extract :: Extract Word32
  extract = build <$> bytes 4

build :: (Bits a, Num a) => ByteString -> a
build = ByteString.foldl' shiftOr 0

shiftOr :: (Bits a1, Num a1) => a1 -> Word8 -> a1
shiftOr word appendix = shiftL word 8 .|. fromIntegral appendix

class GenericExtractable a where
  genericExtract :: Extract (a i)

-- unit data types
instance GenericExtractable U1 where
  genericExtract :: Extract (U1 i)
  genericExtract = pure U1

-- product types
instance GenericExtractable (l :*: r) where
  genericExtract :: Extract ((:*:) l r i)
  genericExtract = genericExtract <|> genericExtract

